import 'dart:typed_data';
import '../coins_base.dart';
import 'package:web3dart/credentials.dart';
import 'package:web3dart/crypto.dart';
import 'package:web3dart/web3dart.dart';
import '../rlp.dart';
import '../wallet_hd_coins.dart';
import 'etransaction.dart';
import 'package:bip39/bip39.dart' as bip39;
import 'package:bitcoin_flutter/bitcoin_flutter.dart' as btcCore;
class CoinsEth extends CoinsBase{

  @override
  String get coinType => "ETH";

  @override
  //这里是18位精度
  int get decimals => 18;

  int get gasLimit=>100000;
  //int get gasLimit=>21000;
  @override
  String get coinInfo => "以太坊（Ethereum）";

  @override
  String get imagePath => "coin/ETH";

  // @override
  // bool get isHaveToken => true;

  @override
  bool get isOpen => true;

  int get gasTokenLimit=>100000;



  @override
  Future<String> getAddress(String subPath) async {
    String pK = await getPrivateKey(subPath);
    EthPrivateKey ethPrivateKey = EthPrivateKey.fromHex(pK);
    EthereumAddress ethAddr = await ethPrivateKey.extractAddress();
    String ethAddress = ethAddr.toString();
    return ethAddress;
  }

  // @override
  // Future<String> getPrivateKey(String subPath) async{
  //   btcCore.HDWallet hdWalletEth =
  //   btcCore.HDWallet.fromSeed(bip39.mnemonicToSeed(getMnemonic))
  //       .derivePath(derivePath(subPath));
  //   String privateKey = hdWalletEth.privKey;
  //   return privateKey;
  // }
  /**
   * num amount, 这个是要发送的币数量，不是wei单位，是小数，也就是真实的eth数量
   * num gasPrice 这个需要的gwei数量，这个是大数，要转成gwei
   */
  @override
  Future<String> transactionMain(String subPath,String fromAddress, String toAddress, num amount, num gasPrice, Map otherPrams) async {
    int nonce  = otherPrams["nonce"];
    Etransaction etransaction  = await _createTransaction(fromAddress,toAddress,amount,gasPrice,nonce);
    String pK = await getPrivateKey(subPath);
    //print("创建交易,chiaid:${network}");

    String txPack = await _signTransaction(pK, etransaction);
    return txPack;
  }

  /**
   * num amount, 这个是要发送的币数量，不是wei单位，是小数，也就是真实的eth数量
   * num gasPrice 这个需要的gwei数量，这个是大数，要转成gwei
   */
  @override
  Future<String> transactionToken(String subPath,String fromAddress, String toAddress, num amount, num gasPrice,String contractAddress, int decimalForToken,Map otherPrams) async {
    int nonce  = otherPrams["nonce"];
    //int dec =decimalForToken;//otherPrams["decimals"];//token的精度往往与主链不一样，所以要传进来
    Etransaction etransaction = await _createErc20Transaction(fromAddress,toAddress,amount,gasPrice,nonce,contractAddress,decimalForToken);
    String pK = await getPrivateKey(subPath);
    String txPack = await _signTransaction(pK, etransaction);
    return txPack;
  }
  // String get derivePath => "m/44'/60'/0'/0/0";
  @override
  String  derivePath(String subPath){
    return "m/44'/60'/$subPath";
  }

  @override
  //正式  1   2021-3-31
  Object get network {
    if(WalletHdCoins.isTestNetWork)
      return 5;
    return 1;
  }
  //测试 3    2021-3-31
  //Object get network => 3;

  // region 辅助方法


  Future<Etransaction> _createTransaction(
      String from, String to, num value, int gasPrice, int nonce) async {

    Transaction transaction = new Transaction(
        from: EthereumAddress.fromHex(from),
        to: EthereumAddress.fromHex(to),
        maxGas: gasLimit,
        gasPrice: EtherAmount.inWei(BigInt.from(gasPrice)),
        value: EtherAmount.inWei(numPow2BigInt(value, decimals)),
        data: hexToBytes(''),
        nonce: nonce);

    return Etransaction(transaction, network);

  }
  Future<String> _signTransaction(
      String privateKeyHex, Etransaction transaction) async {
    if (transaction == null) {
      return null;
    } else {
      EthPrivateKey privateKey = EthPrivateKey.fromHex(privateKeyHex);
      privateKey.extractAddress().then((value) => print(value));
      final Uint8List unsignedTx = uint8ListFromList(encode(_encodeToRlp(
          transaction.tx,
          MsgSignature(BigInt.zero, BigInt.zero, transaction.chainId))));

      final signature = await privateKey.signToSignature(unsignedTx,
          chainId: transaction.chainId);
      final Uint8List signedTx =
      uint8ListFromList(encode(_encodeToRlp(transaction.tx, signature)));

      String signedTxhex = bytesToHex(signedTx);
      return '0x' + signedTxhex;
    }
  }

  /**
   * address：合约地址
   * tokenDecimals :Token 精度
   */
  Future<Etransaction> _createErc20Transaction(
      String from, String to, num value, int gasPrice, int nonce,address //address为合约地址,
      ,int tokenDecimals) async {

    String operationHex =
        '0xa9059cbb000000000000000000000000' + to.toLowerCase().substring(2);
    String valueHex =
    numPow2BigInt(value, tokenDecimals).toRadixString(16);
    String tokenHex = fixhexLength(valueHex);
    String dataHex = operationHex + tokenHex;
//      print(dataHex);
    Transaction transaction = new Transaction(
        from: EthereumAddress.fromHex(from),
        to: EthereumAddress.fromHex(address),
        maxGas: gasTokenLimit,
        gasPrice: EtherAmount.inWei(BigInt.from(gasPrice)),
        value: EtherAmount.zero(),
        data: hexToBytes(dataHex),
        nonce: nonce);

//      print(transaction.gasPrice.getInWei * BigInt.from(transaction.maxGas));
    return Etransaction(transaction, network);

  }

  List<dynamic> _encodeToRlp(
      Transaction transaction, MsgSignature signature) {
    final list = [
      transaction.nonce,
      transaction.gasPrice.getInWei,
      transaction.maxGas,
    ];

    if (transaction.to != null) {
      list.add(transaction.to.addressBytes);
    } else {
      list.add('');
    }

    list..add(transaction.value.getInWei)..add(transaction.data);

    if (signature != null) {
      list..add(signature.v)..add(signature.r)..add(signature.s);
    }
//    print(list);
    return list;
  }

// endregion

}