import 'dart:typed_data';
import '../coins_base.dart';
import '../trx/tronTransactioin.dart';
import 'package:crypto/crypto.dart';
import 'package:web3dart/crypto.dart' as web3dart_crypto;
import 'package:bs58check/bs58check.dart';
import '../wallet_hd_coins.dart';
import 'package:bip39/bip39.dart' as bip39;
import 'package:bitcoin_flutter/bitcoin_flutter.dart' as btcCore;
class CoinsTrx extends CoinsBase{
  @override
  String get coinType => "TRX";
  @override
  int get decimals => 6;
  int get gasLimit=>10000000;
  // @override
  // bool get isHaveToken => true;
  @override
  bool get isOpen => true;
  @override
  Object get network {
    if(WalletHdCoins.isTestNetWork)
      return 3;
    return 1;
  }
  @override
  String get coinInfo => "波场（TRON）";
  @override
  String get imagePath => "coin/TRX";
  @override
  Future<String> getAddress(String subPath) async {
    String pk = await getPrivateKey(subPath);
    Uint8List pubKey = web3dart_crypto.privateKeyBytesToPublic(web3dart_crypto.hexToBytes(pk));
    String pubKeyKeccak256 = web3dart_crypto.bytesToHex(web3dart_crypto.keccak256(pubKey));
    String addressHex = '41' + pubKeyKeccak256.substring(24);
    // print(addressHex);
    Uint8List addressBin = web3dart_crypto.hexToBytes(addressHex);
    String hash1 = sha256.convert(sha256.convert(addressBin).bytes).toString();
    String checksum = hash1.substring(0, 8);
    checksum = addressHex + checksum;
    String addr = base58.encode(web3dart_crypto.hexToBytes(checksum));
    // print(addr);
    return addr;
  }

  @override
  Future<String> transactionMain(String subPath,String fromAddress, String toAddress, num amount, num gasPrice, Map otherPrams) async {
    String refBlockHash = otherPrams["rbh"];
    String refBlockNum = otherPrams["rbn"];
    var createTrxTransaction = await TronTransaction.createTrxTransaction(
        fromAddress,
        toAddress,
        amount,
        refBlockHash,
        refBlockNum,
        gasLimit,decimals);
    String pk = await getPrivateKey(subPath);
    String mSign = await TronTransaction.signTronTransaction(pk,createTrxTransaction);
    return mSign;
  }

  @override
  Future<String> transactionToken(String subPath,String fromAddress, String toAddress, num amount, num gasPrice,String contractAddress, int decimalForToken,Map otherPrams) async {
    String refBlockHash = otherPrams["rbh"];
    String refBlockNum = otherPrams["rbn"];
    //int dec = decimalForToken;//otherPrams["decimals"];//token的精度不能采用主链，所以由归集方传入
    var createTrxTransaction = await TronTransaction.createTrc20Transaction(
        fromAddress,
        toAddress,
        amount,
        refBlockHash,
        refBlockNum,
        gasLimit,
        contractAddress,
        decimalForToken);
    String pk = await getPrivateKey(subPath);
    String mSign = await TronTransaction.signTronTransaction(pk,createTrxTransaction);
    return mSign;
  }
  // String get derivePath => "m/44'/195'/0'/0/0";
  @override
  String  derivePath(String subPath){
    return "m/44'/195'/$subPath";
  }
}