import 'dart:ffi';
import 'dart:typed_data';

// import 'package:crypton/crypton.dart';
import 'package:fixnum/fixnum.dart';
import 'package:crypto/crypto.dart';
import 'package:protobuf/protobuf.dart';
import 'protos/gen/google/protobuf/any.pb.dart';
import 'protos/gen/balance_contract.pb.dart';
import 'protos/gen/Tron.pb.dart';
import 'protos/gen/smart_contract.pb.dart';
import '../rlp.dart';
import 'package:web3dart/crypto.dart' as web3dart_crypto;
import 'package:bs58check/bs58check.dart';

class Ttransaction {
  Uint8List tx;

  Ttransaction(this.tx);
}

class TronTransaction {
  static Future<String> signTronTransaction(
      String privateKeyHex, Ttransaction transaction) async {
    if (transaction == null) {
      return null;
    } else {
      // print(web3dart_crypto.bytesToHex(transaction.tx));
      Transaction tra = Transaction.fromBuffer(transaction.tx);
      Digest txid = sha256.convert(tra.rawData.writeToBuffer());
      // print('txid: ' + txid.toString());

      web3dart_crypto.MsgSignature ms = web3dart_crypto.sign(
          txid.bytes, web3dart_crypto.hexToBytes(privateKeyHex));

      String r = fixhexLength(ms.r.toRadixString(16));
      String s = fixhexLength(ms.s.toRadixString(16));
      String v = fixhexLength((ms.v - 27).toRadixString(16), length: 2);
      String signature = r + s + v;
      // print(r.length);
      // print(r);
      // print(s.length);
      // print(s);
      // print(v.length);
      // print(v);
      // print('signature length: ' + signature.length.toString());
      // print('signature: ' + signature);

      tra.signature.add(web3dart_crypto.hexToBytes(signature));
      Uint8List signedTx = tra.writeToBuffer();
      // print('signedTx: ' + web3dart_crypto.bytesToHex(signedTx));
      return web3dart_crypto.bytesToHex(signedTx);
    }
  }

  static Future<Ttransaction> createTrc20Transaction(String from, String to,
      num value, String refBlockHash, String refBlockNum, int feeLimit, String contractAddress,int decimals) async {
      TriggerSmartContract tsc = TriggerSmartContract.create();
      tsc.ownerAddress = _decode58Check(from);
      tsc.contractAddress = _decode58Check(contractAddress);

      String operationHex = 'a9059cbb000000000000000000000000' +
          web3dart_crypto.bytesToHex(_decode58Check(to)).substring(2);
      String valueHex =
          numPow2BigInt(value, decimals).toRadixString(16);

      String tokenHex = fixhexLength(valueHex);
      String dataHex = operationHex + tokenHex;
      // print('dataHex: ' + dataHex);
      tsc.data = web3dart_crypto.hexToBytes(dataHex);

      Transaction tx = _createTronTransaction(
          tsc,
          Transaction_Contract_ContractType.TriggerSmartContract,
          refBlockHash,
          refBlockNum,
          feeLimit);
      return Ttransaction(tx.writeToBuffer());
    // }
  }

  static Future<Ttransaction> createTrxTransaction(String from, String to,
      num value, String refBlockHash, String refBlockNum, int feeLimit,int decimals) async {
    // if (WalletProxy.tronType.keys.contains(type)) {
    //   CoinInfo coinInfo = WalletProxy.tronType[type];
      TransferContract tfc = TransferContract.create();
      tfc.ownerAddress = _decode58Check(from);
      tfc.toAddress = _decode58Check(to);
      tfc.amount = Int64(numPow2BigInt(value, decimals).toInt());

      Transaction tx = _createTronTransaction(
          tfc,
          Transaction_Contract_ContractType.TransferContract,
          refBlockHash,
          refBlockNum,
          feeLimit);
      return Ttransaction(tx.writeToBuffer());
    // }
  }

  static Uint8List _decode58Check(String input) {
    Uint8List address = base58.decode(input);
    // print(address.length);
    if (address.length == 25) {
      // print(bytesToHex(address.sublist(0, 21)));
      return address.sublist(0, 21);
    }
    return null;
  }

  static Transaction _createTronTransaction(
      GeneratedMessage tfc,
      Transaction_Contract_ContractType ct,
      String refBlockHash,
      String refBlockNum,
      int feeLimit) {
    Transaction tx = Transaction.create();

    Transaction_Contract txc = Transaction_Contract.create();

    try {
      Any any = Any.pack(tfc);
      txc.parameter = any;
    } catch (e) {
      print('TronTransaction.createTronTransaction error : $e');
      return null;
    }
    txc.type = ct;
    Transaction_raw tr = tx.ensureRawData();
    tr.contract.add(txc);
    tr.timestamp = Int64(DateTime.now().millisecondsSinceEpoch);
    tr.expiration = tr.timestamp + 10 * 60 * 60 * 1000;
    print('timestamp: ' + tr.timestamp.toString());
    print('expiration: ' + tr.expiration.toString());

    // tr.timestamp = Int64(1612177284210);
    // tr.expiration = Int64(1612213284210);

    tx.rawData.refBlockHash =
        web3dart_crypto.hexToBytes(refBlockHash.substring(16, 32));
    tx.rawData.refBlockBytes =
        web3dart_crypto.hexToBytes(refBlockHash.substring(12, 16));
    tx.rawData.refBlockNum = Int64.parseInt(refBlockNum);
    tx.rawData.feeLimit = Int64(feeLimit);

    // print(web3dart_crypto.bytesToHex(tx.rawData.refBlockHash));
    // print(web3dart_crypto.bytesToHex(tx.rawData.refBlockBytes));

    return tx;
  }
}
